#ifndef _MONO_BURG_DEFS_
#define _MONO_BURG_DEFS_

/*
 * inssel.brg: instruction selection
 *
 * Author:
 *   Dietmar Maurer (dietmar@ximian.com)
 *   Paolo Molaro (lupus@ximian.com)
 *
 * (C) 2002 Ximian, Inc.
 *
 */
#include <config.h>
#include <string.h>

#include "mini.h"
#include <mono/metadata/marshal.h>
#include <mono/metadata/debug-helpers.h>

#define MBTREE_TYPE  MonoInst
#define MBCGEN_TYPE  MonoCompile
#define MBCOST_DATA  MonoCompile
#define MBALLOC_STATE mono_mempool_alloc (data->state_pool, sizeof (MBState))
#define MBMAX_OPCODES OP_LAST
#define MBGET_OP_NAME(op) mono_inst_name (op)

#define MBTREE_OP(t) ((t)->opcode)
#define MBTREE_LEFT(t) ((t)->inst_left)
#define MBTREE_RIGHT(t) ((t)->inst_right)

#define MONO_EMIT_UNALU(cfg,inst,op,dr,sr1) do { \
                (inst)->opcode = op; \
                (inst)->dreg = dr; \
                (inst)->sreg1 = sr1; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_UNALU(cfg,op,dr,sr1) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
                inst->opcode = op; \
                inst->dreg = dr; \
                inst->sreg1 = sr1; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_BIALU(cfg,inst,op,dr,sr1,sr2) do { \
                (inst)->opcode = op; \
                (inst)->dreg = dr; \
                (inst)->sreg1 = sr1; \
                (inst)->sreg2 = sr2; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_BIALU(cfg,op,dr,sr1,sr2) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
                inst->opcode = op; \
                inst->dreg = dr; \
                inst->sreg1 = sr1; \
                inst->sreg2 = sr2; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_BIALU_IMM(cfg,inst,op,dr,sr,imm) do { \
                (inst)->opcode = op; \
                (inst)->dreg = dr; \
                (inst)->sreg1 = sr; \
                (inst)->inst_p1 = (gpointer)(gssize)(imm); \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_BIALU_IMM(cfg,op,dr,sr,imm) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
                inst->opcode = op; \
                inst->dreg = dr; \
                inst->sreg1 = sr; \
                inst->inst_p1 = (gpointer)(gssize)(imm); \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_BIALU_MEMBASE(cfg,tree,op,dr,sr,basereg,offset) do { \
				tree->opcode = op; \
				tree->dreg = dr; \
				tree->sreg1 = sr; \
				tree->sreg2 = basereg; \
				tree->inst_offset = offset; \
				mono_bblock_add_inst (s->cbb, tree); \
	} while (0)

#define MONO_EMIT_LOAD_MEMBASE(cfg,inst,dr,base,offset) do { \
                (inst)->opcode = OP_LOAD_MEMBASE; \
                (inst)->dreg = dr; \
                (inst)->inst_basereg = base; \
                (inst)->inst_offset = offset; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_LOAD_MEMBASE_OP(cfg,inst,op,dr,base,offset) do { \
                (inst)->opcode = op; \
                (inst)->dreg = dr; \
                (inst)->inst_basereg = base; \
                (inst)->inst_offset = offset; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_LOAD_MEM(cfg,dr,addr) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
                inst->opcode = OP_LOADU4_MEM; \
                inst->dreg = dr; \
                inst->inst_p0 = addr; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_LOAD_MEMBASE(cfg,dr,base,offset) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
                inst->opcode = OP_LOAD_MEMBASE; \
                inst->dreg = dr; \
                inst->inst_basereg = base; \
                inst->inst_offset = offset; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_LOAD_MEMBASE_OP(cfg,op,dr,base,offset) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
                inst->opcode = op; \
                inst->dreg = dr; \
                inst->inst_basereg = base; \
                inst->inst_offset = offset; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_STORE_MEMBASE(cfg,inst,op,base,offset,sr) do { \
                (inst)->opcode = op; \
                (inst)->sreg1 = sr; \
                (inst)->inst_destbasereg = base; \
                (inst)->inst_offset = offset; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_STORE_MEMBASE(cfg,op,base,offset,sr) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
                inst->opcode = op; \
                inst->sreg1 = sr; \
                inst->inst_destbasereg = base; \
                inst->inst_offset = offset; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_STORE_MEMBASE_IMM(cfg,inst,op,base,offset,imm) do { \
                (inst)->opcode = op; \
                (inst)->inst_destbasereg = base; \
                (inst)->inst_offset = offset; \
                (inst)->inst_p1 = (gpointer)imm; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_STORE_MEMBASE_IMM(cfg,op,base,offset,imm) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
                inst->opcode = op; \
                inst->inst_destbasereg = base; \
                inst->inst_offset = offset; \
                inst->inst_p1 = (gpointer)imm; \
	        mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define	MONO_EMIT_NEW_COMPARE_IMM(cfg,sr1,imm) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = OP_COMPARE_IMM;	\
                inst->sreg1 = sr1; \
                inst->inst_p1 = (gpointer)imm; \
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)

#define	MONO_EMIT_NEW_ICOMPARE_IMM(cfg,sr1,imm) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = sizeof (void*) == 8 ? OP_ICOMPARE_IMM : OP_COMPARE_IMM;	\
                inst->sreg1 = sr1; \
                inst->inst_p1 = (gpointer)imm; \
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)

#define	MONO_EMIT_NEW_COND_EXC(cfg,cond,name) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = OP_COND_EXC_##cond;  \
                inst->inst_p1 = (char*)name; \
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)

#define	MONO_EMIT_NEW_ICONST(cfg,dr,imm) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = OP_ICONST;	\
                inst->dreg = dr; \
                inst->inst_c0 = imm; \
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)

#define	MONO_EMIT_NEW_I8CONST(cfg,dr,imm) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = OP_I8CONST;	\
                inst->dreg = dr; \
                inst->inst_l = imm; \
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)

#define	MONO_EMIT_NEW_PCONST(cfg,dr,imm) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = OP_ICONST;	\
                inst->dreg = dr; \
                inst->inst_p0 = (gpointer) imm; \
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)


#ifdef MONO_ARCH_NEED_GOT_VAR

#define MONO_EMIT_NEW_AOTCONST(cfg,dr,cons,patch_type) do { \
		mini_emit_aotconst ((cfg), (dr), (patch_type), (cons)); \
    } while (0)

#else

#define	MONO_EMIT_NEW_AOTCONST(cfg,dr,imm,type) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = OP_AOTCONST;	\
                inst->dreg = dr; \
                inst->inst_p0 = imm; \
                inst->inst_c1 = type; \
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)

#endif

#define	MONO_EMIT_NEW_CLASSCONST(cfg,dr,imm) MONO_EMIT_NEW_AOTCONST(cfg,dr,imm,MONO_PATCH_INFO_CLASS)
#define MONO_EMIT_NEW_VTABLECONST(cfg,dest,vtable) MONO_EMIT_NEW_AOTCONST ((cfg), (dest), (cfg)->compile_aot ? (gpointer)((vtable)->klass) : (vtable), MONO_PATCH_INFO_VTABLE)

#define	MONO_EMIT_NEW_BRANCH_BLOCK(cfg,op,targetbb) do { \
                MonoInst *inst; \
        	MonoInst *target_label; \
		target_label = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		target_label->opcode = OP_LABEL;	\
	        target_label->next = (targetbb)->code; \
		target_label->inst_c0 = (targetbb)->native_offset; \
	        (targetbb)->code = target_label; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = op;	\
		inst->inst_i0 = target_label;	\
		inst->flags = MONO_INST_BRLABEL;	\
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)

#define	MONO_EMIT_NEW_BRANCH_LABEL(cfg,op,label) do { \
                MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = op;	\
		inst->inst_i0 = label;	\
		inst->flags = MONO_INST_BRLABEL;	\
	        mono_bblock_add_inst ((cfg)->cbb, inst); \
	} while (0)

#define MONO_NEW_LABEL(cfg,inst) do { \
		(inst) = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		(inst)->opcode = OP_LABEL;	\
	} while (0)

#define MONO_EMIT_BOUNDS_CHECK(cfg, array_reg, array_type, array_length_field, index_reg) do { \
		if (! (state->tree->flags & MONO_INST_NORANGECHECK)) { \
			int _length_reg = mono_regstate_next_int (cfg->rs); \
			MONO_EMIT_NEW_LOAD_MEMBASE_OP (cfg, OP_LOADI4_MEMBASE, _length_reg, array_reg, G_STRUCT_OFFSET (array_type, array_length_field)); \
			MONO_EMIT_NEW_BIALU (cfg, OP_COMPARE, -1, _length_reg, index_reg); \
			MONO_EMIT_NEW_COND_EXC (cfg, LE_UN, "IndexOutOfRangeException"); \
		} \
	} while (0)

#define MONO_EMIT_BOUNDS_CHECK_IMM(cfg, array_reg, array_type, array_length_field, index_imm) do { \
		if (! (state->tree->flags & MONO_INST_NORANGECHECK)) { \
			int _length_reg = mono_regstate_next_int (cfg->rs); \
			MONO_EMIT_NEW_LOAD_MEMBASE_OP (cfg, OP_LOADI4_MEMBASE, _length_reg, array_reg, G_STRUCT_OFFSET (array_type, array_length_field)); \
			MONO_EMIT_NEW_BIALU_IMM (cfg, OP_COMPARE_IMM, -1, _length_reg, index_imm); \
			MONO_EMIT_NEW_COND_EXC (cfg, LE_UN, "IndexOutOfRangeException"); \
		} \
	} while (0)

/* we need to kludge state because monoburg puts this stuff before the definition of MBState */
void mini_emit_virtual_call (MonoCompile *s, void *state, MonoInst *tree, int novirt_op, int virtop);
void mini_emit_isninst_cast (MonoCompile *s, int klass_reg, MonoClass *klass, MonoInst *false_target, MonoInst *true_target);
void mini_emit_isninst_iface_cast (MonoCompile *s, int vtable_reg, MonoClass *klass, MonoInst *false_target, MonoInst *true_target);
void mini_emit_isninst_iface_class_cast (MonoCompile *s, int klass_reg, MonoClass *klass, MonoInst *false_target, MonoInst *true_target);
void mini_emit_castclass (MonoCompile *s, int klass_reg, MonoClass *klass);
void mini_emit_castclass_iface (MonoCompile *s, int vtable_reg, MonoClass *klass);
void mini_emit_castclass_iface_class (MonoCompile *s, int klass_reg, MonoClass *klass);

void mini_emit_memcpy (MonoCompile *s, int destreg, int doffset, int srcreg, int soffset, int size, int align);
void mini_emit_memset (MonoCompile *s, int destreg, int offset, int size, int val, int align);
int ldind_to_load_membase (int opcode);
void mini_emit_aotconst (MonoCompile *cfg, int dreg, MonoJumpInfoType patch_type, gpointer cons);

#define MONO_EMIT_NEW_X86_COMPARE_MEMBASE_REG(cfg,basereg,offset,operand) do { \
		MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = OP_X86_COMPARE_MEMBASE_REG; \
		inst->inst_basereg = basereg; \
		inst->inst_offset = offset; \
		inst->sreg2 = operand; \
		mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

#define MONO_EMIT_NEW_X86_COMPARE_MEMBASE_IMM(cfg,basereg,offset,operand) do { \
		MonoInst *inst; \
		inst = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
		inst->opcode = OP_X86_COMPARE_MEMBASE_IMM; \
		inst->inst_basereg = basereg; \
		inst->inst_offset = offset; \
		inst->inst_imm = operand; \
		mono_bblock_add_inst (cfg->cbb, inst); \
	} while (0)

/* override the arch independant versions with fast x86 versions */

#undef MONO_EMIT_BOUNDS_CHECK
#undef MONO_EMIT_BOUNDS_CHECK_IMM

#define MONO_EMIT_BOUNDS_CHECK(cfg, array_reg, array_type, array_length_field, index_reg) do { \
		if (! (state->tree->flags & MONO_INST_NORANGECHECK)) { \
			MONO_EMIT_NEW_X86_COMPARE_MEMBASE_REG (cfg, array_reg, G_STRUCT_OFFSET (array_type, array_length_field), index_reg); \
			MONO_EMIT_NEW_COND_EXC (cfg, LE_UN, "IndexOutOfRangeException"); \
		} \
	} while (0)

#define MONO_EMIT_BOUNDS_CHECK_IMM(cfg, array_reg, array_type, array_length_field, index_imm) do { \
		if (! (state->tree->flags & MONO_INST_NORANGECHECK)) { \
			MONO_EMIT_NEW_X86_COMPARE_MEMBASE_IMM (cfg, array_reg, G_STRUCT_OFFSET (array_type, array_length_field), index_imm); \
			MONO_EMIT_NEW_COND_EXC (cfg, LE_UN, "IndexOutOfRangeException"); \
		} \
	} while (0)

int call_reg_to_call_membase (int opcode);
int alu_reg_to_alu_imm (int op);
int alu_reg_to_alu_membase_imm (int op);

#include <glib.h>

#ifndef MBTREE_TYPE
#error MBTREE_TYPE undefined
#endif
#ifndef MBTREE_OP
#define MBTREE_OP(t) ((t)->op)
#endif
#ifndef MBTREE_LEFT
#define MBTREE_LEFT(t) ((t)->left)
#endif
#ifndef MBTREE_RIGHT
#define MBTREE_RIGHT(t) ((t)->right)
#endif
#ifndef MBTREE_STATE
#define MBTREE_STATE(t) ((t)->state)
#endif
#ifndef MBCGEN_TYPE
#define MBCGEN_TYPE int
#endif
#ifndef MBALLOC_STATE
#define MBALLOC_STATE g_new (MBState, 1)
#endif
#ifndef MBCOST_DATA
#define MBCOST_DATA gpointer
#endif

#define MBMAXCOST 32768

#define MBCOND(x) if (!(x)) return MBMAXCOST;

#define MB_TERM_OP_REGVAR	 OP_REGVAR
#define MB_TERM_CEE_LDIND_I	 CEE_LDIND_I
#define MB_TERM_OP_REGOFFSET	 OP_REGOFFSET
#define MB_TERM_OP_LDADDR	 OP_LDADDR
#define MB_TERM_CEE_LDOBJ	 CEE_LDOBJ
#define MB_TERM_OP_ICONST	 OP_ICONST
#define MB_TERM_CEE_ADD	 CEE_ADD
#define MB_TERM_CEE_LDIND_REF	 CEE_LDIND_REF
#define MB_TERM_OP_AOTCONST	 OP_AOTCONST
#define MB_TERM_OP_LOAD_GOTADDR	 OP_LOAD_GOTADDR
#define MB_TERM_CEE_STIND_I	 CEE_STIND_I
#define MB_TERM_OP_PATCH_INFO	 OP_PATCH_INFO
#define MB_TERM_OP_GOT_ENTRY	 OP_GOT_ENTRY
#define MB_TERM_OP_DUMMY_USE	 OP_DUMMY_USE
#define MB_TERM_OP_DUMMY_STORE	 OP_DUMMY_STORE
#define MB_TERM_OP_NOT_REACHED	 OP_NOT_REACHED
#define MB_TERM_OP_MEMORY_BARRIER	 OP_MEMORY_BARRIER
#define MB_TERM_CEE_LDIND_I1	 CEE_LDIND_I1
#define MB_TERM_CEE_LDIND_U1	 CEE_LDIND_U1
#define MB_TERM_CEE_LDIND_I2	 CEE_LDIND_I2
#define MB_TERM_CEE_LDIND_U2	 CEE_LDIND_U2
#define MB_TERM_CEE_LDIND_I4	 CEE_LDIND_I4
#define MB_TERM_CEE_LDIND_U4	 CEE_LDIND_U4
#define MB_TERM_OP_OBJADDR	 OP_OBJADDR
#define MB_TERM_OP_VTADDR	 OP_VTADDR
#define MB_TERM_CEE_STIND_REF	 CEE_STIND_REF
#define MB_TERM_CEE_STIND_I4	 CEE_STIND_I4
#define MB_TERM_CEE_STIND_I1	 CEE_STIND_I1
#define MB_TERM_CEE_STIND_I2	 CEE_STIND_I2
#define MB_TERM_OP_MEMCPY	 OP_MEMCPY
#define MB_TERM_OP_MEMSET	 OP_MEMSET
#define MB_TERM_OP_GROUP	 OP_GROUP
#define MB_TERM_OP_LDELEMA2D	 OP_LDELEMA2D
#define MB_TERM_CEE_CONV_I1	 CEE_CONV_I1
#define MB_TERM_CEE_CONV_I2	 CEE_CONV_I2
#define MB_TERM_CEE_CONV_I4	 CEE_CONV_I4
#define MB_TERM_CEE_CONV_U1	 CEE_CONV_U1
#define MB_TERM_CEE_CONV_U2	 CEE_CONV_U2
#define MB_TERM_CEE_CONV_U4	 CEE_CONV_U4
#define MB_TERM_CEE_CONV_U	 CEE_CONV_U
#define MB_TERM_CEE_CONV_I	 CEE_CONV_I
#define MB_TERM_CEE_CONV_OVF_I4	 CEE_CONV_OVF_I4
#define MB_TERM_CEE_CONV_OVF_U4	 CEE_CONV_OVF_U4
#define MB_TERM_CEE_CONV_OVF_I4_UN	 CEE_CONV_OVF_I4_UN
#define MB_TERM_CEE_CONV_OVF_I1	 CEE_CONV_OVF_I1
#define MB_TERM_CEE_CONV_OVF_I1_UN	 CEE_CONV_OVF_I1_UN
#define MB_TERM_CEE_CONV_OVF_U1	 CEE_CONV_OVF_U1
#define MB_TERM_CEE_CONV_OVF_U1_UN	 CEE_CONV_OVF_U1_UN
#define MB_TERM_CEE_CONV_OVF_I2	 CEE_CONV_OVF_I2
#define MB_TERM_CEE_CONV_OVF_I2_UN	 CEE_CONV_OVF_I2_UN
#define MB_TERM_CEE_CONV_OVF_U2	 CEE_CONV_OVF_U2
#define MB_TERM_CEE_CONV_OVF_U2_UN	 CEE_CONV_OVF_U2_UN
#define MB_TERM_OP_CEQ	 OP_CEQ
#define MB_TERM_OP_CLT	 OP_CLT
#define MB_TERM_OP_CLT_UN	 OP_CLT_UN
#define MB_TERM_OP_CGT	 OP_CGT
#define MB_TERM_OP_CGT_UN	 OP_CGT_UN
#define MB_TERM_OP_LABEL	 OP_LABEL
#define MB_TERM_CEE_NOP	 CEE_NOP
#define MB_TERM_CEE_BREAK	 CEE_BREAK
#define MB_TERM_OP_TLS_GET	 OP_TLS_GET
#define MB_TERM_CEE_SWITCH	 CEE_SWITCH
#define MB_TERM_CEE_BR	 CEE_BR
#define MB_TERM_OP_ARGLIST	 OP_ARGLIST
#define MB_TERM_CEE_CALL	 CEE_CALL
#define MB_TERM_CEE_CALLVIRT	 CEE_CALLVIRT
#define MB_TERM_OP_VOIDCALLVIRT	 OP_VOIDCALLVIRT
#define MB_TERM_OP_FCALLVIRT	 OP_FCALLVIRT
#define MB_TERM_OP_VCALLVIRT	 OP_VCALLVIRT
#define MB_TERM_OP_CALL_REG	 OP_CALL_REG
#define MB_TERM_OP_VOIDCALL	 OP_VOIDCALL
#define MB_TERM_OP_VOIDCALL_REG	 OP_VOIDCALL_REG
#define MB_TERM_OP_FCALL	 OP_FCALL
#define MB_TERM_OP_FCALL_REG	 OP_FCALL_REG
#define MB_TERM_OP_VCALL	 OP_VCALL
#define MB_TERM_OP_VCALL_REG	 OP_VCALL_REG
#define MB_TERM_OP_SAVE_LMF	 OP_SAVE_LMF
#define MB_TERM_OP_RESTORE_LMF	 OP_RESTORE_LMF
#define MB_TERM_OP_I8CONST	 OP_I8CONST
#define MB_TERM_CEE_RET	 CEE_RET
#define MB_TERM_CEE_POP	 CEE_POP
#define MB_TERM_CEE_JMP	 CEE_JMP
#define MB_TERM_CEE_THROW	 CEE_THROW
#define MB_TERM_OP_RETHROW	 OP_RETHROW
#define MB_TERM_OP_THROW_OR_NULL	 OP_THROW_OR_NULL
#define MB_TERM_OP_CALL_HANDLER	 OP_CALL_HANDLER
#define MB_TERM_OP_START_HANDLER	 OP_START_HANDLER
#define MB_TERM_CEE_ENDFINALLY	 CEE_ENDFINALLY
#define MB_TERM_OP_ENDFILTER	 OP_ENDFILTER
#define MB_TERM_OP_CHECK_THIS	 OP_CHECK_THIS
#define MB_TERM_CEE_ISINST	 CEE_ISINST
#define MB_TERM_OP_CISINST	 OP_CISINST
#define MB_TERM_OP_UNBOXCAST	 OP_UNBOXCAST
#define MB_TERM_CEE_CASTCLASS	 CEE_CASTCLASS
#define MB_TERM_OP_CCASTCLASS	 OP_CCASTCLASS
#define MB_TERM_CEE_STELEM_REF	 CEE_STELEM_REF
#define MB_TERM_CEE_NEWARR	 CEE_NEWARR
#define MB_TERM_OP_LMUL	 OP_LMUL
#define MB_TERM_OP_LMUL_OVF	 OP_LMUL_OVF
#define MB_TERM_OP_LMUL_OVF_UN	 OP_LMUL_OVF_UN
#define MB_TERM_OP_LDIV	 OP_LDIV
#define MB_TERM_OP_LDIV_UN	 OP_LDIV_UN
#define MB_TERM_OP_LREM	 OP_LREM
#define MB_TERM_OP_LREM_UN	 OP_LREM_UN
#define MB_TERM_OP_LSHL	 OP_LSHL
#define MB_TERM_OP_LSHR	 OP_LSHR
#define MB_TERM_OP_LSHR_UN	 OP_LSHR_UN
#define MB_TERM_OP_GETCHR	 OP_GETCHR
#define MB_TERM_OP_STR_CHAR_ADDR	 OP_STR_CHAR_ADDR
#define MB_TERM_OP_STRLEN	 OP_STRLEN
#define MB_TERM_OP_GETTYPE	 OP_GETTYPE
#define MB_TERM_OP_GETHASHCODE	 OP_GETHASHCODE
#define MB_TERM_CEE_LDLEN	 CEE_LDLEN
#define MB_TERM_OP_ARRAY_RANK	 OP_ARRAY_RANK
#define MB_TERM_OP_CHECK_ARRAY_TYPE	 OP_CHECK_ARRAY_TYPE
#define MB_TERM_CEE_LDELEMA	 CEE_LDELEMA
#define MB_TERM_CEE_REFANYVAL	 CEE_REFANYVAL
#define MB_TERM_OP_REFANYTYPE	 OP_REFANYTYPE
#define MB_TERM_CEE_MKREFANY	 CEE_MKREFANY
#define MB_TERM_CEE_LDIND_R4	 CEE_LDIND_R4
#define MB_TERM_CEE_LDIND_R8	 CEE_LDIND_R8
#define MB_TERM_CEE_STIND_R4	 CEE_STIND_R4
#define MB_TERM_CEE_STIND_R8	 CEE_STIND_R8
#define MB_TERM_OP_R4CONST	 OP_R4CONST
#define MB_TERM_OP_R8CONST	 OP_R8CONST
#define MB_TERM_OP_FADD	 OP_FADD
#define MB_TERM_OP_FSUB	 OP_FSUB
#define MB_TERM_OP_FMUL	 OP_FMUL
#define MB_TERM_OP_FDIV	 OP_FDIV
#define MB_TERM_OP_FREM	 OP_FREM
#define MB_TERM_OP_FNEG	 OP_FNEG
#define MB_TERM_CEE_CKFINITE	 CEE_CKFINITE
#define MB_TERM_OP_SIN	 OP_SIN
#define MB_TERM_OP_COS	 OP_COS
#define MB_TERM_OP_ABS	 OP_ABS
#define MB_TERM_OP_TAN	 OP_TAN
#define MB_TERM_OP_ATAN	 OP_ATAN
#define MB_TERM_OP_SQRT	 OP_SQRT
#define MB_TERM_OP_FCONV_TO_I4	 OP_FCONV_TO_I4
#define MB_TERM_OP_FCONV_TO_U4	 OP_FCONV_TO_U4
#define MB_TERM_OP_FCONV_TO_OVF_I4	 OP_FCONV_TO_OVF_I4
#define MB_TERM_OP_FCONV_TO_OVF_U4	 OP_FCONV_TO_OVF_U4
#define MB_TERM_OP_FCONV_TO_OVF_I8	 OP_FCONV_TO_OVF_I8
#define MB_TERM_OP_FCONV_TO_OVF_U8	 OP_FCONV_TO_OVF_U8
#define MB_TERM_OP_FCONV_TO_OVF_I	 OP_FCONV_TO_OVF_I
#define MB_TERM_OP_FCONV_TO_OVF_U	 OP_FCONV_TO_OVF_U
#define MB_TERM_OP_FCONV_TO_I	 OP_FCONV_TO_I
#define MB_TERM_OP_FCONV_TO_U	 OP_FCONV_TO_U
#define MB_TERM_OP_FCONV_TO_I2	 OP_FCONV_TO_I2
#define MB_TERM_OP_FCONV_TO_U2	 OP_FCONV_TO_U2
#define MB_TERM_OP_FCONV_TO_I1	 OP_FCONV_TO_I1
#define MB_TERM_OP_FCONV_TO_U1	 OP_FCONV_TO_U1
#define MB_TERM_OP_FCONV_TO_R8	 OP_FCONV_TO_R8
#define MB_TERM_OP_FCONV_TO_R4	 OP_FCONV_TO_R4
#define MB_TERM_CEE_CONV_R4	 CEE_CONV_R4
#define MB_TERM_CEE_CONV_R8	 CEE_CONV_R8
#define MB_TERM_CEE_CONV_R_UN	 CEE_CONV_R_UN
#define MB_TERM_OP_COMPARE	 OP_COMPARE
#define MB_TERM_CEE_AND	 CEE_AND
#define MB_TERM_CEE_OR	 CEE_OR
#define MB_TERM_CEE_XOR	 CEE_XOR
#define MB_TERM_CEE_SUB	 CEE_SUB
#define MB_TERM_CEE_MUL	 CEE_MUL
#define MB_TERM_CEE_MUL_OVF	 CEE_MUL_OVF
#define MB_TERM_CEE_MUL_OVF_UN	 CEE_MUL_OVF_UN
#define MB_TERM_CEE_DIV	 CEE_DIV
#define MB_TERM_CEE_DIV_UN	 CEE_DIV_UN
#define MB_TERM_CEE_REM	 CEE_REM
#define MB_TERM_CEE_REM_UN	 CEE_REM_UN
#define MB_TERM_CEE_SHL	 CEE_SHL
#define MB_TERM_CEE_SHR	 CEE_SHR
#define MB_TERM_CEE_SHR_UN	 CEE_SHR_UN
#define MB_TERM_CEE_NEG	 CEE_NEG
#define MB_TERM_CEE_NOT	 CEE_NOT
#define MB_TERM_CEE_ADD_OVF	 CEE_ADD_OVF
#define MB_TERM_CEE_ADD_OVF_UN	 CEE_ADD_OVF_UN
#define MB_TERM_CEE_SUB_OVF	 CEE_SUB_OVF
#define MB_TERM_CEE_SUB_OVF_UN	 CEE_SUB_OVF_UN
#define MB_TERM_CEE_BNE_UN	 CEE_BNE_UN
#define MB_TERM_CEE_BEQ	 CEE_BEQ
#define MB_TERM_CEE_BLT	 CEE_BLT
#define MB_TERM_CEE_BLT_UN	 CEE_BLT_UN
#define MB_TERM_CEE_BGT	 CEE_BGT
#define MB_TERM_CEE_BGT_UN	 CEE_BGT_UN
#define MB_TERM_CEE_BGE	 CEE_BGE
#define MB_TERM_CEE_BGE_UN	 CEE_BGE_UN
#define MB_TERM_CEE_BLE	 CEE_BLE
#define MB_TERM_CEE_BLE_UN	 CEE_BLE_UN
#define MB_TERM_CEE_CONV_I8	 CEE_CONV_I8
#define MB_TERM_CEE_CONV_U8	 CEE_CONV_U8
#define MB_TERM_CEE_LDIND_I8	 CEE_LDIND_I8
#define MB_TERM_CEE_STIND_I8	 CEE_STIND_I8
#define MB_TERM_OP_BIGMUL	 OP_BIGMUL
#define MB_TERM_OP_BIGMUL_UN	 OP_BIGMUL_UN
#define MB_TERM_OP_LONG_SHRUN_32	 OP_LONG_SHRUN_32
#define MB_TERM_OP_LCONV_TO_I4	 OP_LCONV_TO_I4
#define MB_TERM_OP_LCONV_TO_U4	 OP_LCONV_TO_U4
#define MB_TERM_OP_LADD	 OP_LADD
#define MB_TERM_OP_LADD_OVF	 OP_LADD_OVF
#define MB_TERM_OP_LADD_OVF_UN	 OP_LADD_OVF_UN
#define MB_TERM_OP_LSUB	 OP_LSUB
#define MB_TERM_OP_LSUB_OVF	 OP_LSUB_OVF
#define MB_TERM_OP_LSUB_OVF_UN	 OP_LSUB_OVF_UN
#define MB_TERM_OP_LAND	 OP_LAND
#define MB_TERM_OP_LOR	 OP_LOR
#define MB_TERM_OP_LXOR	 OP_LXOR
#define MB_TERM_OP_LNOT	 OP_LNOT
#define MB_TERM_OP_LNEG	 OP_LNEG
#define MB_TERM_OP_LCOMPARE	 OP_LCOMPARE
#define MB_TERM_CEE_CONV_OVF_U8	 CEE_CONV_OVF_U8
#define MB_TERM_CEE_CONV_OVF_I8	 CEE_CONV_OVF_I8
#define MB_TERM_CEE_CONV_OVF_I8_UN	 CEE_CONV_OVF_I8_UN
#define MB_TERM_CEE_CONV_OVF_U8_UN	 CEE_CONV_OVF_U8_UN
#define MB_TERM_OP_LCONV_TO_R_UN	 OP_LCONV_TO_R_UN
#define MB_TERM_OP_FCONV_TO_I8	 OP_FCONV_TO_I8
#define MB_TERM_OP_FCONV_TO_U8	 OP_FCONV_TO_U8
#define MB_TERM_OP_LCONV_TO_U8	 OP_LCONV_TO_U8
#define MB_TERM_OP_LCONV_TO_I8	 OP_LCONV_TO_I8
#define MB_TERM_OP_LCONV_TO_U	 OP_LCONV_TO_U
#define MB_TERM_OP_LCONV_TO_I	 OP_LCONV_TO_I
#define MB_TERM_OP_LCONV_TO_I1	 OP_LCONV_TO_I1
#define MB_TERM_OP_LCONV_TO_U1	 OP_LCONV_TO_U1
#define MB_TERM_OP_LCONV_TO_I2	 OP_LCONV_TO_I2
#define MB_TERM_OP_LCONV_TO_U2	 OP_LCONV_TO_U2
#define MB_TERM_OP_LCONV_TO_OVF_I1_UN	 OP_LCONV_TO_OVF_I1_UN
#define MB_TERM_OP_LCONV_TO_OVF_I1	 OP_LCONV_TO_OVF_I1
#define MB_TERM_OP_LCONV_TO_OVF_U1_UN	 OP_LCONV_TO_OVF_U1_UN
#define MB_TERM_OP_LCONV_TO_OVF_U1	 OP_LCONV_TO_OVF_U1
#define MB_TERM_OP_LCONV_TO_OVF_I2	 OP_LCONV_TO_OVF_I2
#define MB_TERM_OP_LCONV_TO_OVF_I2_UN	 OP_LCONV_TO_OVF_I2_UN
#define MB_TERM_OP_LCONV_TO_OVF_U2_UN	 OP_LCONV_TO_OVF_U2_UN
#define MB_TERM_OP_LCONV_TO_OVF_U2	 OP_LCONV_TO_OVF_U2
#define MB_TERM_OP_LCONV_TO_OVF_U4_UN	 OP_LCONV_TO_OVF_U4_UN
#define MB_TERM_OP_LCONV_TO_OVF_I_UN	 OP_LCONV_TO_OVF_I_UN
#define MB_TERM_OP_LCONV_TO_OVF_U4	 OP_LCONV_TO_OVF_U4
#define MB_TERM_OP_LCONV_TO_OVF_I	 OP_LCONV_TO_OVF_I
#define MB_TERM_OP_LCONV_TO_OVF_I4_UN	 OP_LCONV_TO_OVF_I4_UN
#define MB_TERM_OP_LCONV_TO_OVF_I4	 OP_LCONV_TO_OVF_I4
#define MB_TERM_OP_LCONV_TO_OVF_I8_UN	 OP_LCONV_TO_OVF_I8_UN
#define MB_TERM_OP_LCONV_TO_OVF_U8	 OP_LCONV_TO_OVF_U8
#define MB_TERM_OP_LCONV_TO_OVF_I8	 OP_LCONV_TO_OVF_I8
#define MB_TERM_OP_LCONV_TO_OVF_U8_UN	 OP_LCONV_TO_OVF_U8_UN
#define MB_TERM_OP_LCALLVIRT	 OP_LCALLVIRT
#define MB_TERM_OP_LCALL	 OP_LCALL
#define MB_TERM_OP_LCALL_REG	 OP_LCALL_REG
#define MB_TERM_OP_LCONV_TO_R8	 OP_LCONV_TO_R8
#define MB_TERM_OP_LCONV_TO_R4	 OP_LCONV_TO_R4
#define MB_TERM_OP_LOCALLOC	 OP_LOCALLOC
#define MB_TERM_OP_SETRET	 OP_SETRET
#define MB_TERM_OP_OUTARG	 OP_OUTARG
#define MB_TERM_OP_OUTARG_R4	 OP_OUTARG_R4
#define MB_TERM_OP_OUTARG_R8	 OP_OUTARG_R8
#define MB_TERM_OP_OUTARG_VT	 OP_OUTARG_VT
#define MB_TERM_OP_X86_OUTARG_ALIGN_STACK	 OP_X86_OUTARG_ALIGN_STACK
#define MB_TERM_OP_ATOMIC_ADD_NEW_I4	 OP_ATOMIC_ADD_NEW_I4
#define MB_TERM_OP_ATOMIC_ADD_I4	 OP_ATOMIC_ADD_I4
#define MB_TERM_OP_ATOMIC_EXCHANGE_I4	 OP_ATOMIC_EXCHANGE_I4

#define MB_NTERM_stmt	1
#define MB_NTERM_reg	2
#define MB_NTERM_base	3
#define MB_NTERM_cflags	4
#define MB_NTERM_freg	5
#define MB_NTERM_lreg	6
#define MB_NTERM_fpcflags	7
#define MB_NTERM_i8con	8
#define MB_MAX_NTERMS	8

typedef struct _MBState MBState;
struct _MBState {
	int		 op;
	MBTREE_TYPE	 *tree;
	gint32 reg1, reg2;
	MBState		*left, *right;
	guint16		cost[9];
	unsigned int	 rule_stmt:8;
	unsigned int	 rule_reg:8;
	unsigned int	 rule_base:4;
	unsigned int	 rule_cflags:6;
	unsigned int	 rule_freg:6;
	unsigned int	 rule_lreg:6;
	unsigned int	 rule_fpcflags:1;
	unsigned int	 rule_i8con:2;
};

typedef void (*MBEmitFunc) (MBState *state, MBTREE_TYPE *tree, MBCGEN_TYPE *s);

extern const char * const mono_burg_term_string [];
#if MONOBURG_LOG
extern const char * const mono_burg_rule_string [];
#endif /* MONOBURG_LOG */
extern const guint16 mono_burg_nts_data [];
extern const guint8 mono_burg_nts [];
extern MBEmitFunc const mono_burg_func [];
MBState *mono_burg_label (MBTREE_TYPE *tree, MBCOST_DATA *data);
int mono_burg_rule (MBState *state, int goal);
MBState **mono_burg_kids (MBState *state, int rulenr, MBState *kids []);
extern void mono_burg_init (void);

#endif /* _MONO_BURG_DEFS_ */
